from typing import Iterable

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.keymetadata import KeyMetadata, KeyMetadataFactory
from rainbowadn.data.collection.trees.binary.activebinarytree import ActiveBinaryTree
from rainbowadn.data.collection.trees.binary.avl import AVLBTBP
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree, BinaryTreeFactory
from rainbowadn.data.collection.trees.comparison.comparator import Fail
from rainbowadn.data.collection.trees.comparison.hashcomparator import HashComparator
from rainbowadn.hashing.hash_point_format import tabulate
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullable import Nullable
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable
from rainbowadn.hashing.resolverorigin import ResolverOrigin
from rainbowadn.hashing.static import StaticFactory, StaticMentionable
from rainbowadn.v13.algo import MINT_CONST
from rainbowadn.v13.subject import Subject
from rainbowadn.v13.transaction import Coin, Transaction

__all__ = ('BankState',)


class BankState(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            minted: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            used: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            miner: NullableReference[Subject],
            length: HashPoint[Integer]
    ):
        assert isinstance(miner, NullableReference)
        assert isinstance(used, NullableReference)
        assert isinstance(miner, NullableReference)
        assert isinstance(length, HashPoint)
        self.minted = minted
        self.used = used
        self.miner = miner
        self.length = length

    def points(self) -> Iterable[HashPoint]:
        return [*self.minted.points(), *self.used.points(), *self.miner.points(), self.length]

    def __bytes__(self):
        return bytes(self.minted) + bytes(self.used) + bytes(self.miner) + bytes(self.length)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'BankState':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        reference_factory: RainbowFactory[
            NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]]
        ] = NullableReferenceFactory(
            BinaryTreeFactory(
                KeyMetadataFactory(
                    StaticFactory(Coin),
                    StaticFactory(Integer),
                ).loose()
            ).loose()
        ).loose()
        assert isinstance(reference_factory, RainbowFactory)
        return cls(
            reference_factory.from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            reference_factory.from_bytes(source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver),
            NullableReferenceFactory(
                StaticFactory(Subject)
            ).from_bytes(source[2 * HashPoint.HASH_LENGTH:3 * HashPoint.HASH_LENGTH], resolver),
            ResolverOrigin(Integer.factory(), source[3 * HashPoint.HASH_LENGTH:], resolver).hash_point()
        )

    def without_miner(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            self.length
        )

    def advance(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            HashPoint.of(Integer(self.length.resolve().integer + 1))
        )

    def minted_tree(self) -> ActiveBinaryTree[Coin, Integer]:
        return ActiveBinaryTree(
            AVLBTBP(HashComparator(Fail())), self.minted
        )

    def used_tree(self) -> ActiveBinaryTree[Coin, Integer]:
        return ActiveBinaryTree(
            AVLBTBP(HashComparator(Fail())), self.used
        )

    def use_coins(self, coins: Iterable[HashPoint[Coin]]) -> 'BankState':
        minted: ActiveBinaryTree[Coin, Integer] = self.minted_tree()
        used: ActiveBinaryTree[Coin, Integer] = self.used_tree()
        for in_coin in coins:
            assert minted.contains(in_coin)
            assert not used.contains(in_coin)
            used = used.add(in_coin)
        return BankState(self.minted, used.reference, self.miner, self.length)

    def mint_coins(
            self,
            transaction: Transaction
    ) -> 'BankState':
        assert self.verify(transaction)
        miner = self.miner_nullable()
        minted: ActiveBinaryTree[Coin, Integer] = self.minted_tree()
        for coin, miner in transaction.iter_coins(self.mint(), miner):
            assert not minted.contains(HashPoint.of(coin))
            minted = minted.add(HashPoint.of(coin))
            assert isinstance(minted, ActiveBinaryTree)
        return BankState(minted.reference, self.used, NullableReference(miner, self.miner.factory), self.length)

    def miner_nullable(self) -> Nullable[HashPoint[Subject]]:
        return self.miner.reference

    def mint(self) -> int:
        if self.miner_nullable().null():
            return MINT_CONST
        else:
            return 0

    def verify(self, transaction: Transaction) -> bool:
        assert isinstance(transaction, Transaction)
        assert transaction.verify(self.mint())
        return True

    def _push(self, transaction: Transaction) -> 'BankState':
        assert isinstance(transaction, Transaction)
        return self.use_coins(
            transaction.data.resolve().iter_in_coins()
        ).mint_coins(
            transaction
        )

    def push(self, transaction: HashPoint[Transaction]) -> 'BankState':
        return self._push(transaction.resolve())

    def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}bank' \
               f'{tabulate(tab + 1)}(miner)' \
               f'{tabulate(tab + 1)}{self.miner.str(tab + 1)}' \
               f'{tabulate(tab + 1)}(minted)' \
               f'{tabulate(tab + 1)}{self.minted.str(tab + 1)}' \
               f'{tabulate(tab + 1)}(used)' \
               f'{tabulate(tab + 1)}{self.used.str(tab + 1)}' \
               f'{tabulate(tab + 1)}(length)' \
               f'{tabulate(tab + 1)}{self.length.resolve()}' \
               f'{tabulate(tab)})'
